"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const module_1 = require("module");
const path_1 = require("path");
const enhanced_resolve_1 = __importDefault(require("enhanced-resolve"));
const create = enhanced_resolve_1.default.create;
exports.default = ({ types: t }, options) => {
    const { ignoreBuiltins, ignoreBareImport, ignore, relative: optionRelative = true, cache = true, ..._resolveOptions } = options;
    const ignores = ignore || [];
    if (ignoreBuiltins) {
        ignores.push(...module_1.builtinModules);
    }
    if (ignoreBareImport) {
        ignores.push(/^[^\.\/]/);
    }
    if (!cache) {
        _resolveOptions.fileSystem = fs_1.default;
    }
    const resolveOptions = {
        conditionNames: ["import"],
        extensions: [".tsx", ".ts", ".jsx", ".mjs", ".js", ".cjs", ".json", ".node"],
        ..._resolveOptions,
    };
    const enhancedResolve = create.sync(resolveOptions);
    const resolve = (path, request, deps) => {
        if (ignores.some((ignore) => (typeof ignore === "string" ? ignore === request : ignore.test(request)))) {
            deps.push(request);
            return request;
        }
        const dir = path_1.dirname(path);
        request = enhancedResolve(dir, request) || request;
        deps.push(request);
        if (optionRelative) {
            request = path_1.relative(dir, request);
            if (request[0] !== "/" && request[0] !== ".") {
                request = "./" + request;
            }
        }
        return request;
    };
    const transformDeclaration = (path, state) => {
        const { source } = path.node;
        if (!state.filename || !source) {
            return;
        }
        let modulePath = source.value;
        try {
            modulePath = resolve(state.filename, modulePath, state.file.metadata.deps);
        }
        catch (e) {
            console.log(path.buildCodeFrameError(e?.message));
            return;
        }
        const sourcePath = path.get("source");
        if (Array.isArray(sourcePath)) {
            return;
        }
        sourcePath.replaceWith(t.stringLiteral(modulePath));
    };
    const transformImport = (path, state) => {
        if (!state.filename || !path.parentPath.isCallExpression()) {
            return;
        }
        const arg = path.parentPath.get("arguments.0");
        if (Array.isArray(arg)) {
            return;
        }
        let modulePath;
        if (arg.isTemplateLiteral() && arg.node.expressions.length === 0) {
            modulePath = arg.node.quasis[0].value.cooked;
        }
        if (arg.isStringLiteral()) {
            modulePath = arg.node.value;
        }
        if (!modulePath) {
            return;
        }
        try {
            modulePath = resolve(state.filename, modulePath, state.file.metadata.deps);
        }
        catch (e) {
            console.log(path.buildCodeFrameError(e?.message));
            return;
        }
        arg.replaceWith(t.stringLiteral(modulePath));
    };
    return {
        name: "enhanced-resolve",
        pre(_state) {
            this.file.metadata.deps = [];
        },
        post() {
            this.file.metadata.deps = [...new Set(this.file.metadata.deps.sort())];
        },
        visitor: {
            ImportDeclaration: transformDeclaration,
            ExportAllDeclaration: transformDeclaration,
            ExportNamedDeclaration: transformDeclaration,
            Import: transformImport,
        },
    };
};
//# sourceMappingURL=index.js.map